/**
 *
 * $Header: /cvsroot/lesstif/lesstif/lib/Xm-2.0/Container.c,v 1.38 2001/10/23 22:50:34 dannybackx Exp $
 *
 * Copyright (C) 1997-1999 Free Software Foundation, Inc.
 * Copyright  1999, 2000, 2001 LessTif Development Team
 *
 * This file is part of the GNU LessTif Library.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **/
/* 
 * MODULE:	Container.c
 * DESCRIPTION:	Implement the XmContainer widget class.
 * STATUS:	Not too well yet.
 * 
 */

#include <LTconfig.h>

#include <XmI/XmI.h>
#include <Xm/XmP.h>
#include <Xm/BaseClassP.h>
#include <Xm/Container.h>
#include <Xm/ContainerP.h>
#include <Xm/ContainerT.h>
#include <Xm/ContItemT.h>
#include <Xm/TransltnsP.h>
#include <Xm/PushBG.h>

#include <XmI/DebugUtil.h>

static const char rcsid[] = "$Header: /cvsroot/lesstif/lesstif/lib/Xm-2.0/Container.c,v 1.38 2001/10/23 22:50:34 dannybackx Exp $";

/*
 * Forward Declarations
 */

static void class_initialize(void);
static void class_part_initialize(WidgetClass widget_class);
static void Initialize(Widget, Widget, ArgList, Cardinal *);
/* static void resize(Widget w); */
static Boolean SetValues(Widget, Widget, Widget, ArgList, Cardinal *);
static void expose(Widget w, XEvent *event, Region region);
static XtGeometryResult QueryGeometry(Widget, XtWidgetGeometry *, XtWidgetGeometry *);
static void _XmContainerLayout(Widget w, Dimension *pwid, Dimension *pht);

/* composite */
static XtGeometryResult GeometryManager(Widget, XtWidgetGeometry *, XtWidgetGeometry *);
static void ChangeManaged(Widget w);

/* constraint */
static void ConstraintInitialize(Widget request, Widget new_w, ArgList args, Cardinal *num_args);
static void ConstraintDestroy(Widget w);
static Boolean ConstraintSetValues(Widget, Widget, Widget, ArgList, Cardinal *);

/*
 * Actions
 */
static void ContainerActivate(Widget, XEvent *, String *, Cardinal *);
static void ContainerBeginExtend(Widget, XEvent *, String *, Cardinal *);
static void ContainerBeginSelect(Widget, XEvent *, String *, Cardinal *);
static void ContainerBeginToggle(Widget, XEvent *, String *, Cardinal *);
static void ContainerButtonMotion(Widget, XEvent *, String *, Cardinal *);
static void ContainerCancel(Widget, XEvent *, String *, Cardinal *);
static void ContainerDeselectAll(Widget, XEvent *, String *, Cardinal *);
static void ContainerEndExtend(Widget, XEvent *, String *, Cardinal *);
static void ContainerEndSelect(Widget, XEvent *, String *, Cardinal *);
static void ContainerEndToggle(Widget, XEvent *, String *, Cardinal *);
static void ContainerEndTransfer(Widget, XEvent *, String *, Cardinal *);
static void ContainerExpandOrCollapse(Widget, XEvent *, String *, Cardinal *);
static void ContainerExtend(Widget, XEvent *, String *, Cardinal *);
static void ContainerExtendCursor(Widget, XEvent *, String *, Cardinal *);
static void ContainerHandleBtn1Down(Widget, XEvent *, String *, Cardinal *);
static void ContainerHandleBtn1Motion(Widget, XEvent *, String *, Cardinal *);
static void ContainerHandleBtn1Up(Widget, XEvent *, String *, Cardinal *);
static void ContainerHandleBtn2Down(Widget, XEvent *, String *, Cardinal *);
static void ContainerHandleBtn2Motion(Widget, XEvent *, String *, Cardinal *);
static void ContainerHandleBtn2Up(Widget, XEvent *, String *, Cardinal *);
static void ContainerMoveCursor(Widget, XEvent *, String *, Cardinal *);
static void ContainerPrimaryCopy(Widget, XEvent *, String *, Cardinal *);
static void ContainerPrimaryLink(Widget, XEvent *, String *, Cardinal *);
static void ContainerPrimaryMove(Widget, XEvent *, String *, Cardinal *);
static void ContainerSelect(Widget, XEvent *, String *, Cardinal *);
static void ContainerSelectAll(Widget, XEvent *, String *, Cardinal *);
static void ContainerStartTransfer(Widget, XEvent *, String *, Cardinal *);
static void ContainerToggleMode(Widget, XEvent *, String *, Cardinal *);

/*
 * Trait
 */
static void _ContainerTrait_GetValues(Widget container, XmContainerData d);

static XmContainerTraitRec _ContainerTraitRec = {
	/* version */		0,
	/* getvalues */		_ContainerTrait_GetValues
};

/*
 * Bitmaps, "stolen" from XltListTree
 */
#define folder_width 16
#define folder_height 12
static unsigned char folder_bits[] =
{
  0x00, 0x1f, 0x80, 0x20, 0x7c, 0x5f, 0x02, 0x40, 0x02, 0x40, 0x02, 0x40,
  0x02, 0x40, 0x02, 0x40, 0x02, 0x40, 0x02, 0x40, 0x02, 0x40, 0xfc, 0x3f,
};

#define folderopen_width 16
#define folderopen_height 12
static unsigned char folderopen_bits[] =
{
  0x00, 0x3e, 0x00, 0x41, 0xf8, 0xd5, 0xac, 0xaa, 0x54, 0xd5, 0xfe, 0xaf,
  0x01, 0xd0, 0x02, 0xa0, 0x02, 0xe0, 0x04, 0xc0, 0x04, 0xc0, 0xf8, 0x7f,
};

enum {
	_XmUnknown,
	_XmIconButton,
	_XmChild
};

/*
 * resources for the Container class
 */
#define Offset(field) XtOffsetOf(XmContainerRec, container.field)
static XtResource resources[] =
{
    {
	XmNmarginHeight, XmCMarginHeight, XmRVerticalDimension,
	sizeof(Dimension), Offset(margin_h),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNmarginWidth, XmCMarginWidth, XmRHorizontalDimension,
	sizeof(Dimension), Offset(margin_w),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNautomaticSelection, XmCAutomaticSelection, XmRChar,
	sizeof(unsigned char), Offset(automatic),
	XmRImmediate, (XtPointer)XmAUTO_SELECT
    },
    {
	XmNcollapsedStatePixmap, XmCCollapsedStatePixmap, XmRPixmap,
	sizeof(Pixmap), Offset(collapsed_state_pixmap),
	XmRImmediate, (XtPointer)XmUNSPECIFIED_PIXMAP	/* FIXME default should be dynamic */
    },
    {
	XmNconvertCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(convert_cb),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNdefaultActionCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(default_action_cb),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNdestinationCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(destination_cb),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNdetailColumnHeading, XmCDetailColumnHeading, XmRXmStringTable,
	sizeof(XmStringTable), Offset(detail_heading),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNdetailColumnHeadingCount, XmCDetailColumnHeadingCount, XmRCardinal,
	sizeof(Cardinal), Offset(detail_heading_count),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNdetailOrder, XmCDetailOrder, XmRCardinalList,
	sizeof(Cardinal *), Offset(detail_order),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNdetailOrderCount, XmCDetailOrderCount, XmRCardinal,
	sizeof(Cardinal), Offset(detail_order_count),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNdetailTabList, XmCDetailTabList, XmRTabList,
	sizeof(XmTabList), Offset(detail_tablist),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNentryViewType, XmCEntryViewType, XmRChar,
	sizeof(unsigned char), Offset(entry_viewtype),
	XmRImmediate, (XtPointer)XmANY_ICON
    },
    {
	XmNexpandedStatePixmap, XmCExpandedStatePixmap, XmRPixmap,
	sizeof(Pixmap), Offset(expanded_state_pixmap),
	XmRImmediate, (XtPointer)XmUNSPECIFIED_PIXMAP	/* FIXME should be dynamic */
    },
#if 0
    {
	XmNfontList, XmCFontList, XmRFontList,
	sizeof(XmFontList), Offset(fontlist),
	XmRImmediate, (XtPointer)0
    },
#endif
    {
	XmNlargeCellHeight, XmCCellHeight, XmRDimension,
	sizeof(Dimension), Offset(large_cell_height),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNlargeCellWidth, XmCCellWidth, XmRDimension,
	sizeof(Dimension), Offset(large_cell_width),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNlayoutType, XmCLayoutType, XmRChar,
	sizeof(unsigned char), Offset(layout_type),
	XmRImmediate, (XtPointer)XmSPATIAL
    },
    {
	XmNoutlineButtonPolicy, XmCOutlineButtonPolicy, XmRChar,
	sizeof(unsigned char), Offset(ob_policy),
	XmRImmediate, (XtPointer)XmOUTLINE_BUTTON_PRESENT
    },
    {
	XmNoutlineChangedCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(outline_cb),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNoutlineColumnWidth, XmCOutlineColumnWidth, XmRDimension,
	sizeof(Dimension), Offset(ob_width),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNoutlineIndentation, XmCOutlineIndentation, XmRDimension,
	sizeof(Dimension), Offset(outline_indent),
	XmRImmediate, (XtPointer)40
    },
    {
	XmNoutlineLineStyle, XmCLineStyle, XmRChar,
	sizeof(unsigned char), Offset(outline_sep_style),
	XmRImmediate, (XtPointer)XmSINGLE
    },
    {
	XmNprimaryOwnership, XmCPrimaryOwnership, XmRChar,
	sizeof(unsigned char), Offset(primary_ownership),
	XmRImmediate, (XtPointer)XmOWN_POSSIBLE_MULTIPLE
    },
    {
	XmNrenderTable, XmCRenderTable, XmRRenderTable,
	sizeof(XmRenderTable), Offset(render_table),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNselectColor, XmCSelectColor, XmRPixel,
	sizeof(Pixel), Offset(select_color),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNselectedObjects, XmCSelectedObjects, XmRWidgetList,
	sizeof(WidgetList), Offset(selected_items),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNselectedObjectCount, XmCSelectedObjectCount, XmRInt,
	sizeof(unsigned int), Offset(selected_item_count),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNselectionCallback, XmCCallback, XmRCallback,
	sizeof(XtCallbackList), Offset(selection_cb),
	XmRImmediate, (XtPointer)0
    },
    {
	XmNselectionPolicy, XmCSelectionPolicy, XmRChar,
	sizeof(unsigned char), Offset(selection_policy),
	XmRImmediate, (XtPointer)XmEXTENDED_SELECT
    },
    {
	XmNselectionTechnique, XmCSelectionTechnique, XmRChar,
	sizeof(unsigned char), Offset(selection_technique),
	XmRImmediate, (XtPointer)XmTOUCH_OVER
    },
    {
	XmNsmallCellHeight, XmCCellHeight, XmRDimension,
	sizeof(Dimension), Offset(small_cell_height),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNsmallCellWidth, XmCCellWidth, XmRDimension,
	sizeof(Dimension), Offset(small_cell_width),
	XmRImmediate, (XtPointer)0	/* FIXME should be dynamic */
    },
    {
	XmNspatialIncludeModel, XmCSpatialIncludeModel, XmRChar,
	sizeof(unsigned char), Offset(include_model),
	XmRImmediate, (XtPointer)XmAPPEND
    },
    {
	XmNspatialResizeModel, XmCSpatialResizeModel, XmRChar,
	sizeof(unsigned char), Offset(resize_model),
	XmRImmediate, (XtPointer)XmGROW_MINOR
    },
    {
	XmNspatialSnapModel, XmCSpatialSnapModel, XmRChar,
	sizeof(unsigned char), Offset(snap_model),
	XmRImmediate, (XtPointer)XmNONE
    },
    {
	XmNspatialStyle, XmCSpatialStyle, XmRChar,
	sizeof(unsigned char), Offset(spatial_style),
	XmRImmediate, (XtPointer)XmGRID
    },
#undef	Offset
};

#define	Offset(x) (XtOffsetOf(XmContainerConstraintRec, container.x))
static XtResource containerConstraintResources [] = {
	{
		XmNentryParent, XmCWidget, XmRWidget,
		sizeof(Widget), Offset(entry_parent),
		XmRImmediate, (XtPointer)0
	},
	{
		XmNoutlineState, XmCOutlineState, XmRUnsignedChar,
		sizeof(unsigned char), Offset(outline_state),
		XmRImmediate, XmCOLLAPSED
	},
	{
		XmNpositionIndex, XmCPositionIndex, XmRInt,
		sizeof(int), Offset(position_index),
		XmRImmediate, (XtPointer)0	/* FIX ME init in ConstraintInitialize ? */
	},
};
#undef Offset

static XtActionsRec actions[] = {
	{ "ContainerActivate",		ContainerActivate },
	{ "ContainerBeginExtend",	ContainerBeginExtend },
	{ "ContainerBeginSelect",	ContainerBeginSelect },
	{ "ContainerBeginToggle",	ContainerBeginToggle },
	{ "ContainerButtonMotion",	ContainerButtonMotion },
	{ "ContainerCancel",		ContainerCancel },
	{ "ContainerDeselectAll",	ContainerDeselectAll },
	{ "ContainerEndExtend",		ContainerEndExtend },
	{ "ContainerEndSelect",		ContainerEndSelect },
	{ "ContainerEndToggle",		ContainerEndToggle },
	{ "ContainerEndTransfer",	ContainerEndTransfer },
	{ "ContainerExpandOrCollapse",	ContainerExpandOrCollapse },
	{ "ContainerExtend",		ContainerExtend },
	{ "ContainerExtendCursor",	ContainerExtendCursor },
	{ "ContainerHandleBtn1Down",	ContainerHandleBtn1Down },
	{ "ContainerHandleBtn1Motion",	ContainerHandleBtn1Motion },
	{ "ContainerHandleBtn1Up",	ContainerHandleBtn1Up },
	{ "ContainerHandleBtn2Down",	ContainerHandleBtn2Down },
	{ "ContainerHandleBtn2Motion",	ContainerHandleBtn2Motion },
	{ "ContainerHandleBtn2Up",	ContainerHandleBtn2Up },
	{ "ContainerMoveCursor",	ContainerMoveCursor },
	{ "ContainerPrimaryCopy",	ContainerPrimaryCopy },
	{ "ContainerPrimaryLink",	ContainerPrimaryLink },
	{ "ContainerPrimaryMove",	ContainerPrimaryMove },
	{ "ContainerSelect",		ContainerSelect },
	{ "ContainerSelectAll",		ContainerSelectAll },
	{ "ContainerStartTransfer",	ContainerStartTransfer },
	{ "ContainerToggleMode",	ContainerToggleMode },
};

/* *INDENT-OFF* */
XmContainerClassRec xmContainerClassRec = {
    /* Core class part */
    {
	/* superclass            */ (WidgetClass) &xmManagerClassRec,
        /* class_name            */ "XmContainer",
	/* widget_size           */ sizeof(XmContainerRec),
	/* class_initialize      */ class_initialize,
	/* class_part_initialize */ class_part_initialize,
	/* class_inited          */ False,
	/* initialize            */ Initialize,
	/* initialize_hook       */ NULL,
	/* realize               */ XtInheritRealize,
	/* actions               */ actions,
	/* num_actions           */ XtNumber(actions),
	/* resources             */ resources,
	/* num_resources         */ XtNumber(resources),
	/* xrm_class             */ NULLQUARK,
	/* compress_motion       */ True,
	/* compress_exposure     */ XtExposeCompressMaximal,
	/* compress_enterleave   */ True,
	/* visible_interest      */ False,
	/* destroy               */ NULL,
	/* resize                */ XtInheritResize,
	/* expose                */ expose,
	/* set_values            */ SetValues,
	/* set_values_hook       */ NULL,
	/* set_values_almost     */ XtInheritSetValuesAlmost,
	/* get_values_hook       */ NULL,
	/* accept_focus          */ NULL,
	/* version               */ XtVersion,
	/* callback offsets      */ NULL,
	/* tm_table              */ _XmContainer_defaultTranslations,
	/* query_geometry        */ QueryGeometry,
	/* display_accelerator   */ NULL,
	/* extension             */ (XtPointer)NULL /*&_XmContainerCoreClassExtRec*/
    },
    /* Composite class part */
    {
	/* geometry manager */ GeometryManager, 
        /* change_managed   */ ChangeManaged, 
        /* insert_child     */ XtInheritInsertChild,
        /* delete_child     */ XtInheritDeleteChild,
        /* extension        */ (XtPointer)NULL
    },
    /* Constraint class part */
    {
	/* subresources      */ containerConstraintResources,
        /* subresource_count */ XtNumber(containerConstraintResources), 
        /* constraint_size   */ sizeof(XmContainerConstraintRec),
        /* initialize        */ ConstraintInitialize,
        /* destroy           */ ConstraintDestroy,
        /* set_values        */ ConstraintSetValues,
        /* extension         */ NULL,  
    },
    /* XmManager class part */
    {
	/* translations                 */	XtInheritTranslations,
        /* syn_resources                */	NULL,
        /* num_syn_resources            */	0,
        /* syn_constraint_resources     */	NULL,
        /* num_syn_constraint_resources */	0,
        /* parent_process               */	XmInheritParentProcess,
	/* extension                    */	(XtPointer)NULL
    },
    /* XmContainer Area part */
    {
	/* extension */				NULL,
    },
};
/* *INDENT-ON* */

WidgetClass xmContainerWidgetClass = (WidgetClass)&xmContainerClassRec;

Widget
XmCreateContainer(Widget parent, String name, ArgList arglist, Cardinal argcount)
{
	return XtCreateWidget(name, xmContainerWidgetClass, parent, arglist, argcount);
}

static void class_part_initialize(WidgetClass widget_class)
{
	/* FIX ME: Needs to be implemented */
	DEBUGOUT(_LtDebug(__FILE__, NULL, "XmContainer class_part_initialize\n"));
}

static XtTranslations container_traversal_trans;

static void
class_initialize(void)
{
	container_traversal_trans = XtParseTranslationTable(_XmContainer_traversalTranslations);

	if (! XmeTraitSet((XtPointer)xmContainerWidgetClass, XmQTcontainer,
			(XtPointer) &_ContainerTraitRec)) {
		_XmWarning(NULL, "XmContainer ClassInitialize: XmeTraitSet failed\n");
	}
}

static void
Initialize(Widget request, Widget new_w, ArgList args, Cardinal *num_args)
{
	Dimension	wid, ht;
	XGCValues	values;
	XtGCMask	mask;

	/* FIX ME: Needs to be implemented */

	DEBUGOUT(_LtDebug(__FILE__, new_w, "XmContainer initialize\n"));

	XtAugmentTranslations(new_w, container_traversal_trans);

	ContainerFirstNode(new_w) = 0;
	ContainerLastNode(new_w) = 0;
	ContainerIconHeader(new_w) = 0;

	_XmContainerLayout(new_w, &wid, &ht);

	XtWidth(new_w) = (wid < 100) ? 100 : wid;
	XtHeight(new_w) = (ht < 100) ? 100 : ht;

	if (ContainerCollapsedStatePixmap(new_w) == XmUNSPECIFIED_PIXMAP) {
		ContainerCollapsedStatePixmap(new_w) = XCreateBitmapFromData(
			XtDisplay(new_w), RootWindowOfScreen(XtScreen(new_w)),
			(char *)folder_bits, folder_width, folder_height);
	}
	if (ContainerExpandedStatePixmap(new_w) == XmUNSPECIFIED_PIXMAP) {
		ContainerExpandedStatePixmap(new_w) = XCreateBitmapFromData(
			XtDisplay(new_w), RootWindowOfScreen(XtScreen(new_w)),
			(char *)folderopen_bits, folderopen_width, folderopen_height);
	}

	ContainerMarqueeGC(new_w) = NULL;	/* FIX ME */
	mask = GCForeground;
	values.foreground = MGR_Foreground(new_w);
	ContainerNormalGC(new_w) = XtGetGC(new_w, mask, &values);
}

#if 0
/* FIX ME Need this ? */
static void
resize(Widget w)
{
	/* FIX ME: Needs to be implemented */
	DEBUGOUT(_LtDebug(__FILE__, w, "XmContainer resize\n"));
}
#endif

static Boolean
SetValues(Widget current, Widget request, Widget new_w, ArgList args, Cardinal *num_args)
{
	/* FIX ME: Needs to be implemented */
#define	NE(r)	(r(request) != r(new_w))

	DEBUGOUT(_LtDebug(__FILE__, new_w, "XmContainer SetValues\n"));
	DEBUGOUT(_LtDebugPrintArgList(__FILE__, new_w, args, *num_args, False));

	if (NE(ContainerEntryViewType)) {
		/* Must pass this to children using the ContainerItem trait */

		XmContainerItemTrait	tr;
		int			i;
		Widget			child;
		XmContainerItemDataRec	data;

		data.valueMask = ContItemViewType;
		data.view_type = ContainerEntryViewType(new_w);

		for (i=0; i<MGR_NumChildren(new_w); i++) {
			child = MGR_Children(new_w)[i];
			if ((tr = (XmContainerItemTrait)XmeTraitGet((XtPointer)XtClass(child),
								XmQTcontainerItem))) {
				tr->setValues(child, &data);
			} else {
				continue;	/* Shouldn't happen */
			}
		}
	}

	return True;
#undef	NE
}

static void
expose(Widget w, XEvent *event, Region region)
{
	if (! XtIsRealized(w))
		return;

	DEBUGOUT(_LtDebug(__FILE__, w, "XmContainer expose\n"));
	/* FIX ME: Needs to be implemented */

	_XmRedisplayGadgets(w, event, region);

	if (ContainerLayoutType(w) == XmOUTLINE) { /* Draw the lines */
		int	i;

		for (i=0; i<MGR_NumChildren(w); i++) {
			Widget 		c, last, child = MGR_Children(w)[i];
			CwidNode	cn, n;
			Position	x0, y0, my;

			if (CC_OutlineState(child) == XmCOLLAPSED)
				continue;
			if ((cn = CC_Child(child)) == NULL)
				continue;

			/* Find the non-icon child with the highest Y */
			my=-1; last = NULL;
			for (n=cn; n && n->next_ptr; n=n->next_ptr) {
				c = n->widget_ptr;

				if (CC_CWidType(c) == _XmIconButton)
					continue;
				else if (XtY(c) > my) {
					my = XtY(c);
					last = c;
				}
			}
			if (! last)
				last = cn->widget_ptr;

			my = XtY(last) + XtHeight(last) / 2;
			x0 = XtX(child) + XtWidth(child) / 2;
			y0 = XtY(child) + XtHeight(child);

			/* Vertical line */
			XDrawLine(XtDisplay(w), XtWindow(w), ContainerNormalGC(w),
				x0, y0, x0, my);
			/*
			 * Horizontal lines
			 * Draw these only to the folder button if a child has children of its own.
			 */
			for (n=cn; n; n=n->next_ptr) {
				Position	x, y;

				c = n->widget_ptr;

				if (CC_CWidType(c) == _XmIconButton)
					continue;

				y = XtY(c) + XtHeight(c) / 2;

				if (n->child_ptr && CC_Related(c)) {
					Widget	r = CC_Related(c);
					x = XtX(r);
				} else {
					x = XtX(c);
				}

				XDrawLine(XtDisplay(w), XtWindow(w), ContainerNormalGC(w),
					x0, y, x, y);
			}
			
		}
	}
}

/*
 * Recursive function to walk down the tree
 */
static void
_XmContainerLayoutOL(Widget w, CwidNode n, Position x0, Position y0,
			Dimension *pwid, Dimension *pht, int lvl)
{
	Position	oldx = x0,
			mx, my, x, y;
	CwidNode	node;

/* FIX ME */
#define	HMARGIN	4
#define	WMARGIN	4

	if (n == NULL)
		return;

	/*
	 * If we have a lvl parameter then we might as well use it to
	 * prevent us from having infinite recursion.
	 */
	if (lvl > MGR_NumChildren(w) + 5) {
		_XmWarning(w, "Recursion problem !!\n");
		return;
	}

	mx = x0;
	my = y0;
	node = n;
	while (node) {
		Widget	nw = node->widget_ptr;

		x0 = oldx;

		DEBUGOUT(_LtDebug2(__FILE__, w, nw,
			"_XmContainerLayoutOL[%d](X0 %d Y0 %d)(next %s, child %s)\n",
			lvl, x0, y0,
			node->next_ptr ? XtName(node->next_ptr->widget_ptr)
					: "none",
			node->child_ptr ? XtName(node->child_ptr->widget_ptr)
					: "none"));

		if (CC_CWidType(nw) == _XmIconButton) {
			node = node->next_ptr;
			continue;
		}

		/* Position the widget and, if relevant, its icon */
		if (node->child_ptr && CC_Related(nw)) {
			Widget other = CC_Related(nw);
			Position	ox, oy;

			ox = x0;
			oy = y0;
			x = x0 + WMARGIN + ContainerMarginWidth(w) + XtWidth(other);
			y = y0;

			if (ox != XtX(other) || oy != XtY(other)) {
				_XmMoveObject(other, ox, oy);
				DEBUGOUT(_LtDebug2(__FILE__, w, other,
					"Move related child to %d %d\n", ox, oy));
			}
			if (x != XtX(nw) || y != XtY(nw)) {
				_XmMoveObject(nw, x, y);
			}
			mx = x + XtWidth(nw) + XtWidth(other)
				+ 2 * ContainerMarginWidth(w) + 2 * WMARGIN;
			my = y + ContainerMarginHeight(w) + XtHeight(nw) + HMARGIN;
		} else {
			x = x0;
			y = y0;
			if (x != XtX(nw) || y != XtY(nw)) {
				_XmMoveObject(nw, x, y);
			}
			
			mx = x + ContainerMarginWidth(w) + WMARGIN + XtWidth(nw);
			my = y + ContainerMarginHeight(w) + HMARGIN + XtHeight(nw);
		}
		DEBUGOUT(_LtDebug2(__FILE__, w, nw,
			"Move child to %d %d\n", x, y));

		if (node->child_ptr) {
			if (CC_OutlineState(nw) == XmEXPANDED) {
				_XmContainerLayoutOL(w, node->child_ptr,
					mx, my,
					&x0, &y0, lvl+1);
				if (x0 > mx)
					mx = x0;
				if (my < y0)
					my = y0;
			} else { /* XmCOLLAPSED */
				/* FIX ME  x0 = mx; */ y0 = my;
			}
		}
		node = node->next_ptr;

		/* X0 is updated at the beginning of the loop. */
		y0 = my;
	}
	x0 = mx;
	if (*pwid < mx)
		*pwid = mx;
	y0 = my;
	if (*pht < my)
		*pht = my;
	DEBUGOUT(_LtDebug(__FILE__, w, "_XmContainerLayoutOL[%d] -> %d %d\n",
		lvl, x0, y0));
}

/*
 * FIX ME this is simplistic
 *
 * Show items in a tree, in PositionIndex order within EntryParent order.
 * Lines are drawn to show parent-child relationship.
 */
static void
_XmContainerLayoutOutline(Widget w, Dimension *pwid, Dimension *pht)
{
	Dimension wid, ht;

	wid = ContainerMarginWidth(w);
	ht = 0;

	DEBUGOUT(_LtDebug(__FILE__, w, "_XmContainerLayoutOutline(%d children)\n",
		MGR_NumChildren(w)));

	_XmContainerLayoutOL(w, ContainerFirstNode(w),
		ContainerMarginWidth(w), ContainerMarginHeight(w),
		&wid, &ht, 0);
	if (pwid)
		*pwid = wid;
	if (pht)
		*pht = ht;
	return;
}

/*
 * FIX ME this is simplistic
 *
 * ContainerLayoutType == XmSPATIAL
 *
 * Layout influenced by
 *	ContainerSpatialStyle (XmCELLS, XmGRID, XmNONE)
 *	ContainerSpatialIncludeModel (XmAPPEND, XmCLOSEST, XmFIRST_FIT)
 *	ContainerSpatialSnapModel (XmCENTER, XmSNAP_TO_GRID, XmNONE)
 */
static void
_XmContainerLayoutSpatial(Widget w, Dimension *pwid, Dimension *pht)
{
	Dimension wid, ht;
	Cardinal i;
	Position x, y;

	wid = ContainerMarginWidth(w);
	ht = 0;

	for (i=0; i<MGR_NumChildren(w); i++) {
		Widget	child = MGR_Children(w)[i];
		if (XtIsManaged(child)) {
			wid += XtWidth(child) + ContainerMarginWidth(w);
			if (XtHeight(child) > ht)
				ht = XtHeight(child);
		}
	}

	ht += 2 * ContainerMarginHeight(w);

	if (pwid)
		*pwid = wid;
	if (pht)
		*pht = ht;

	x = ContainerMarginWidth(w);
	y = ContainerMarginHeight(w);

	for (i=0; i<MGR_NumChildren(w); i++) {
		Widget	child = MGR_Children(w)[i];
		if (XtIsManaged(child)) {
			_XmMoveObject(child, x, y);
			x += XtWidth(child) + ContainerMarginWidth(w);

			if (x > XtWidth(w) - ContainerMarginWidth(w)) {
				x = ContainerMarginWidth(w);
				if (XtHeight(child) < 20)
					y += 30;
				y += XtHeight(child);	/* FIX ME simplistic */
			}
		}
	}
}

/*
 * FIX ME this is simplistic
 */
static void
_XmContainerLayoutSimple(Widget w, Dimension *pwid, Dimension *pht)
{
	Dimension wid, ht;
	Cardinal i;
	Position x, y;

	wid = ContainerMarginWidth(w);
	ht = 0;

	for (i=0; i<MGR_NumChildren(w); i++) {
		Widget	child = MGR_Children(w)[i];
		if (XtIsManaged(child)) {
			wid += XtWidth(child) + ContainerMarginWidth(w);
			if (XtHeight(child) > ht)
				ht = XtHeight(child);
		}
	}

	ht += 2 * ContainerMarginHeight(w);

	if (pwid)
		*pwid = wid;
	if (pht)
		*pht = ht;

	x = ContainerMarginWidth(w);
	y = ContainerMarginHeight(w);

	for (i=0; i<MGR_NumChildren(w); i++) {
		Widget	child = MGR_Children(w)[i];
		if (XtIsManaged(child)) {
			_XmMoveObject(child, x, y);
			x += XtWidth(child) + ContainerMarginWidth(w);

			if (x > XtWidth(w) - ContainerMarginWidth(w)) {
				x = ContainerMarginWidth(w);
				if (XtHeight(child) < 20)
					y += 30;
				y += XtHeight(child);	/* FIX ME simplistic */
			}
		}
	}
}

/*
 * FIX ME this is simplistic
 */
static void
_XmContainerLayoutDetail(Widget w, Dimension *pwid, Dimension *pht)
{
	Dimension wid, ht;
	Cardinal i;
	Position x, y;

	wid = ContainerMarginWidth(w);
	ht = 0;

	for (i=0; i<MGR_NumChildren(w); i++) {
		Widget	child = MGR_Children(w)[i];
		if (XtIsManaged(child)) {
			wid += XtWidth(child) + ContainerMarginWidth(w);
			if (XtHeight(child) > ht)
				ht = XtHeight(child);
		}
	}

	ht += 2 * ContainerMarginHeight(w);

	if (pwid)
		*pwid = wid;
	if (pht)
		*pht = ht;

	x = ContainerMarginWidth(w);
	y = ContainerMarginHeight(w);

	for (i=0; i<MGR_NumChildren(w); i++) {
		Widget	child = MGR_Children(w)[i];
		if (XtIsManaged(child)) {
			_XmMoveObject(child, x, y);
			x += XtWidth(child) + ContainerMarginWidth(w);

			if (x > XtWidth(w) - ContainerMarginWidth(w)) {
				x = ContainerMarginWidth(w);
				if (XtHeight(child) < 20)
					y += 30;
				y += XtHeight(child);	/* FIX ME simplistic */
			}
		}
	}
}
static void
_XmContainerLayout(Widget w, Dimension *pwid, Dimension *pht)
{
	switch (ContainerLayoutType(w)) {
	case XmDETAIL:
		_XmContainerLayoutDetail(w, pwid, pht);
		break;
	case XmOUTLINE:
		_XmContainerLayoutOutline(w, pwid, pht);
		break;
	case XmSPATIAL:
		_XmContainerLayoutSpatial(w, pwid, pht);
		break;
	default:
		_XmContainerLayoutSimple(w, pwid, pht);
		break;
	}
}

static XtGeometryResult
QueryGeometry(Widget w, XtWidgetGeometry *request, XtWidgetGeometry *reply)
{
	Dimension		wid, ht;
	XtWidgetGeometry	g;

	/* FIX ME: Needs to be implemented */

	DEBUGOUT(_LtDebug(__FILE__, w, "XmContainer QueryGeometry\n"));
	g = *request;

	_XmContainerLayout(w, &wid, &ht);

	wid = (request->request_mode & CWWidth) ?
		(request->width < wid) ? wid : request->width : wid;
	ht = (request->request_mode & CWHeight) ?
		(request->height < ht) ? ht : request->height : ht;

	reply->request_mode = CWWidth | CWHeight;
	reply->width = (wid == 0) ? 1 : wid;
	reply->height = (ht == 0) ? 1 : ht;

	return _XmGMReplyToQueryGeometry(w, &g, reply);
}

/* composite */
static XtGeometryResult
GeometryManager(Widget w, XtWidgetGeometry *request, XtWidgetGeometry *reply)
{
	DEBUGOUT(_LtDebug(__FILE__, w, "XmContainer GeometryManager\n"));

	/* FIX ME: Needs to be implemented */
	*reply = *request;
	return XtGeometryYes;
}

static void
ChangeManaged(Widget w)
{
	DEBUGOUT(_LtDebug(__FILE__, w, "XmContainer ChangeManaged\n"));

	/* FIX ME: Needs to be implemented */
	_XmNavigChangeManaged(w);
	_XmContainerLayout(w, 0, 0);
}

static void
_XmContainerExpandNode(Widget c)
{
	CwidNode	n;
	Widget		r;

	CC_OutlineState(c) = XmEXPANDED;

	r = CC_Related(c);
	if (r)
		XtVaSetValues(r,
			XmNoutlineState, XmEXPANDED,
			XmNlabelPixmap, ContainerExpandedStatePixmap(XtParent(r)),
			NULL);

	for (n=CC_Node(c)->child_ptr; n; n=n->next_ptr)
		XtManageChild(n->widget_ptr);
}

static void
_XmContainerUnmapSubtree(CwidNode n)
{
	CwidNode	c;
	Widget		w, r;

	DEBUGOUT(_LtDebug(__FILE__, n->widget_ptr, "_XmContainerUnmapSubtree\n"));
	if (!n)
		return;

	w = n->widget_ptr;
	if (CC_CWidType(w) == _XmIconButton)
		return;

	CC_OutlineState(w) = XmCOLLAPSED;
	r = CC_Related(w);
	if (r)
		XtVaSetValues(r,
			XmNoutlineState, XmCOLLAPSED,
			XmNlabelPixmap, ContainerCollapsedStatePixmap(XtParent(r)),
			NULL);

	for (c=n->child_ptr; c; c=c->next_ptr)
		if (CC_CWidType(c->widget_ptr) != _XmIconButton) {
			XtUnmanageChild(c->widget_ptr);
			_XmContainerUnmapSubtree(c);
		}
}

/*
 * The pushbutton gadget gets toggled; need to open/close this part
 * of the tree. "Client" is the associated IconGadget.
 */
static void
_XmContainerToggle(Widget w, XtPointer client, XtPointer call)
{
	Widget	b = (Widget) client,
		o = CC_Related(b);
	Dimension	wid, ht;

	DEBUGOUT(_LtDebug2(__FILE__, w, o, "_XmContainerToggle\n"));

	if (CC_OutlineState(o) == XmCOLLAPSED) {
		_XmContainerExpandNode(o);
	} else {
		/* Unmap children and their children */
		_XmContainerUnmapSubtree(CC_Node(o));
	}
	_XmContainerLayout(w, &wid, &ht);

	XtWidth(w) = wid;
	XtHeight(w) = ht;

	XClearWindow(XtDisplay(w), XtWindow(w));
}

/* constraint */
/* FIX ME must maintain ContainerLastNode */
static void
ConstraintInitialize(Widget request, Widget new_w, ArgList args, Cardinal *num_args)
{
	Widget				cw = XtParent(new_w), prev = NULL;
/*	XmContainerConstraintPtr	con = (XmContainerConstraintPtr)CoreConstraints(new_w); */
	Dimension			wid, ht;

	DEBUGOUT(_LtDebug2(__FILE__, cw, new_w, "ConstraintInitialize\n"));
	DEBUGOUT(
		if (CC_EntryParent(new_w)) {
		_LtDebug2(__FILE__, new_w, CC_EntryParent(new_w),
			"EntryParent\n");
	});

	/* Need to initialize PositionIndex. FIX ME */
	if (CC_PositionIndex(new_w) == 0) {
		int	i, m;

		/* Find children with the same EntryParent */
		/* Get the maximum PositionIndex among them */
		m = -1;
		for (i=0; i<MGR_NumChildren(cw); i++) {
			Widget c = MGR_Children(cw)[i];

			if (CC_CWidType(c) == _XmIconButton)
				continue;

			if (CC_EntryParent(c) == CC_EntryParent(new_w) && CC_PositionIndex(c) > m) {
				m = CC_PositionIndex(c);
				prev = c;
			}
		}

		/* Assign one more than the maximum found */
		if (m >= 0) {
			CC_PositionIndex(new_w) = m + 1;
			DEBUGOUT(_LtDebug2(__FILE__, new_w, prev,
				"positionIndex %d\n", m+1));
		}
	}

	if (CC_OutlineState(new_w) == 0xFF) {
		/* This is the value we pass when creating a "folder" icon */
		DEBUGOUT(_LtDebug2(__FILE__, cw, new_w, "Yes !\n"));
		CC_CWidType(new_w) = _XmIconButton;
	} else {
		CC_CWidType(new_w) = _XmChild;
	}

	/* Build the tree structure if OUTLINE */
	CC_Node(new_w) = NULL;
	CC_Related(new_w) = NULL;

	if (ContainerLayoutType(cw) == XmOUTLINE) {
		CC_Node(new_w) = (CwidNode)XtMalloc(sizeof(XmCwidNodeRec));
		CC_Next(new_w) = CC_Prev(new_w) = CC_Child(new_w) = NULL;

		CC_Widget(new_w) = new_w;
		
		if (CC_EntryParent(new_w)) {
			CC_Parent(new_w) = CC_Node(CC_EntryParent(new_w));
			DEBUGOUT(_LtDebug2(__FILE__, cw, new_w,
				"ConstraintInit: parent is %s %p\n",
				XtName(CC_EntryParent(new_w)),
				CC_EntryParent(new_w)));
		} else {
			CC_Parent(new_w) = 0;
		}

		if (! prev) {
			/* FIX ME look for the previous node */
		}
		if (prev) {
			if (CC_Next(prev)) {	/* Need to do some fixing */
				/* Now : prev -> nxt , to be prev -> this -> nxt */
				Widget	nxt = CC_Next(prev)->widget_ptr;
				CC_Prev(new_w) = CC_Node(prev);
				CC_Next(new_w) = CC_Node(nxt);
				CC_Prev(nxt) = CC_Node(new_w);
				CC_Next(prev) = CC_Node(new_w);
			}
			CC_Prev(new_w) = CC_Node(prev);
			CC_Next(prev) = CC_Node(new_w);
		}
	}

	/* First/last */
	if (ContainerFirstNode(cw) == 0) {
		ContainerFirstNode(cw) = CC_Node(new_w);
		ContainerLastNode(cw) = CC_Node(new_w);
	}

	DEBUGOUT(_LtDebug2(__FILE__, cw, new_w,
		"XmContainer ConstraintInitialize (parent %p PI %d X %d Y %d)\n",
		CC_EntryParent(new_w),
		CC_PositionIndex(new_w),
		CC_UserX(new_w),
		CC_UserY(new_w)));

	/* Create a button for my parent if it doesn't have any yet */
	if (ContainerLayoutType(cw) == XmOUTLINE
			&& CC_CWidType(new_w) == _XmChild
			&& CC_Node(new_w)
			&& CC_Parent(new_w)) {
		Widget	parent = CC_Parent(new_w)->widget_ptr;

		if (!CC_Related(parent)) {
	 		Arg	al[5];
			int	ac = 0;
			Widget	b;

			if (CC_OutlineState(parent) == XmCOLLAPSED) {
				XtSetArg(al[ac], XmNlabelPixmap, ContainerCollapsedStatePixmap(cw)); ac++;
			} else {
				XtSetArg(al[ac], XmNlabelPixmap, ContainerExpandedStatePixmap(cw)); ac++;
			}
			XtSetArg(al[ac], XmNlabelType, XmPIXMAP); ac++;
			XtSetArg(al[ac], XmNoutlineState, (XtPointer)0xFF); ac++;
			CC_Related(parent) =
			b = XmCreatePushButtonGadget(cw, "icon", al, ac);
			CC_Related(b) = parent;
			XtManageChild(b);
			CC_CWidType(b) = _XmIconButton;

			/*
			 * Allocate a node for the new button,
			 * and insert it in the tree.
			 * This is mucho easier here which is why we
			 * pull the 0xFF outlineState trick above.
			 */
			CC_Node(b) = (CwidNode)XtMalloc(sizeof(XmCwidNodeRec));
			CC_Next(b) = CC_Next(new_w);
			CC_Prev(b) = CC_Node(new_w);
			CC_Next(new_w) = CC_Node(b);
			if (CC_Next(b))
				CC_Next(b)->prev_ptr = CC_Node(b);
			CC_Child(b) = NULL;
			CC_Parent(b) = CC_Node(parent);

			CC_Widget(b) = b;

			/* Doesn't currently seem to work FIX ME */
			XtAddCallback(b, XmNactivateCallback,
				_XmContainerToggle, (XtPointer)parent);
		}

		/* Obviously this means the parent didn't have a child yet */
		if (CC_Child(parent) == NULL)
			CC_Child(parent) = CC_Node(new_w);
	}

	_XmContainerLayout(cw, &wid, &ht);
	if (XtWidth(cw) < wid)
		XtWidth(cw) = wid;
	if (XtHeight(cw) < ht)
		XtHeight(cw) = ht;
}

static void
ConstraintDestroy(Widget w)
{
	Widget	cw = XtParent(w);

	/* FIX ME: Needs to be implemented */
	DEBUGOUT(_LtDebug2(__FILE__, cw, w, "XmContainer ConstraintDestroy\n"));
}

static Boolean
ConstraintSetValues(Widget current, Widget request, Widget new_w, ArgList args, Cardinal *num_args)
{
	Widget	c = XtParent(new_w);
	Boolean	relayout = False;

	DEBUGOUT(_LtDebug2(__FILE__, c, new_w, "XmContainer ConstraintSetValues\n"));
	/* FIX ME: Needs to be implemented */
#if 0
	if (CC_OutlineState(new_w) != CC_OutlineState(current)) {
		Pixmap	p;

		if (CC_OutlineState(new_w) == XmCOLLAPSED)
			p = ContainerCollapsedStatePixmap(c);
		else
			p = ContainerExpandedStatePixmap(c);

		XtVaSetValues(new_w, XmNlabelPixmap, p, NULL);
	}
#endif
	if (relayout)
		_XmContainerLayout(c, NULL, NULL);
	return relayout;
}


int
XmContainerGetItemChildren(Widget w, Widget item, WidgetList *item_children)
{
	_XmWarning(w, "XmContainerGetItemChildren(): not yet implemented!");
	return 0;
}

void
XmContainerRelayout(Widget w)
{
	_XmWarning(w, "XmContainerRelayout(): not yet implemented!");
}

void
XmContainerReorder(Widget w, WidgetList cwid_list, int cwid_count)
{
	_XmWarning(w, "XmContainerReorder(): not yet implemented!");
}

Boolean
XmContainerCut(Widget w, Time timestamp)
{
	_XmWarning(NULL, "XmContainerCut(): not yet implemented!");
	return False;
}

Boolean
XmContainerCopy(Widget w, Time timestamp)
{
	_XmWarning(w, "XmContainerCopy(): not yet implemented!");
	return False;
}


Boolean XmContainerPaste(Widget w)
{
	_XmWarning(w, "XmContainerPaste(): not yet implemented!");
	return False;
}


Boolean XmContainerCopyLink(Widget w, Time timestamp)
{
	_XmWarning(w, "XmContainerCopyLink(): not yet implemented!");
	return False;
}


Boolean XmContainerPasteLink(Widget w)
{
	_XmWarning(w, "XmContainerPasteLink(): not yet implemented!");
	return False;
}

/*
 * typedef struct {
 * 	Mask		valueMask;
 * 	Cardinal	*detail_order;
 * 	Cardinal	detail_order_count;
 * 	XmTabList	detail_tablist;
 * 	Dimension	first_column_width;
 * 	unsigned char	selection_mode;
 * 	Pixel		select_color;
 * } XmContainerDataRec, *XmContainerData;
 */
static void
_ContainerTrait_GetValues(Widget container, XmContainerData d)
{
	DEBUGOUT(_LtDebug(__FILE__, container, "_ContainerTrait_GetValues\n"));
}

/*
 * Figure out which object the user 'hit'
 */
static Widget
HitObject(Widget cont, Position x, Position y)
{
	/* FIX ME this is currently very simplistic */
	int	i;

	for (i=0; i<MGR_NumChildren(cont); i++) {
		Widget child = MGR_Children(cont)[i];

		if (XtX(child) <= x && x <= XtX(child) + XtWidth(child)
			&& XtY(child) <= y && y <= XtY(child) + XtHeight(child)) {

			return child;
		}
	}
	return NULL;
}

/*
 * Actions
 */
static void
ContainerActivate(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	XmContainerSelectCallbackStruct	cbs;
	Widget				fw;
#if 0
	DEBUGOUT(_LtDebugAction(__FILE__, w, "XmContainer ContainerActivate", params, num_params));
#endif
	/* This should probably be the technique to get this ...
	fw = XmGetFocusWidget(w);
	** */

	fw = ContainerIconHeader(w);	/* Hack */
	DEBUGOUT(_LtDebug2(__FILE__, w, fw, "ContainerActivate\n"));

	if (CC_CWidType(fw) == _XmIconButton) {
		_XmContainerToggle(w, (XtPointer)fw, NULL);
	} else {
		/* */
		cbs.reason = XmCR_DEFAULT_ACTION;
		cbs.event = event;
		cbs.selected_items = NULL;	/* FIX ME */
		cbs.selected_item_count = 0;	/* FIX ME */
		cbs.auto_selection_type = 0;	/* FIX ME */

		XtCallCallbackList(w, ContainerDefaultActionCallback(w), (XtPointer)&cbs);
	}
}


static void
ContainerBeginExtend(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerBeginExtend", params, num_params));
}


static void
ContainerBeginSelect(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	Time	mctime = XtGetMultiClickTime(XtDisplay(w));
	Widget	hit;

	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerBeginSelect", params, num_params));
	DEBUGOUT(_LtDebug(__FILE__, w, "\tEvent at %d %d\n", event->xbutton.x, event->xbutton.y));

	/* If this is the second ContainerBeginSelect within the multiclick time,
	 * then call the default action callback */
	if (ContainerLastClickTime(w) &&
			event->xbutton.time - ContainerLastClickTime(w) < mctime) {
		XmContainerSelectCallbackStruct	cbs;

		cbs.reason = XmCR_DEFAULT_ACTION;
		cbs.event = event;
		cbs.selected_items = NULL;	/* FIX ME */
		cbs.selected_item_count = 0;	/* FIX ME */
		cbs.auto_selection_type = 0;	/* FIX ME */

		XtCallCallbackList(w, ContainerDefaultActionCallback(w), (XtPointer)&cbs);
		return;
	}

	ContainerLastClickTime(w) = event->xbutton.time;
	hit = HitObject(w, event->xbutton.x, event->xbutton.y);
	ContainerIconHeader(w) = hit;	/* FIX ME this resource probably serves another puspose */

	DEBUGOUT(_LtDebug(__FILE__, w, "Hit '%s'\n",
		hit ? XtName(hit) : "(nothing)"));

	/* FIX ME another hack */
	if (CC_CWidType(hit) == _XmIconButton) {
#if 1
		/* It's probably not as simple as this */
		XtCallActionProc(w, "ManagerGadgetActivate", event, NULL, 0);
#endif
	}

	switch (ContainerSelectionPolicy(w)) {
	case XmSINGLE_SELECT:
		if (ContainerSelectedItems(w)) {
			XtFree((char *)ContainerSelectedItems(w));
		}
		ContainerSelectedItems(w) = NULL;
		ContainerSelectedItemCount(w) = 0;
		if (hit) {
			ContainerSelectedItems(w) = (WidgetList)XtCalloc(sizeof(Widget *), 1);
			ContainerSelectedItems(w)[0] = hit;
			ContainerSelectedItemCount(w) = 1;
		}
		break;
	case XmBROWSE_SELECT:
	case XmMULTIPLE_SELECT:
	case XmEXTENDED_SELECT:
	default:
	}
}

static void
ContainerBeginToggle(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerBeginToggle", params, num_params));
}

static void
ContainerButtonMotion(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerButtonMotion", params, num_params));
}

static void
ContainerCancel(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerCancel", params, num_params));
}

static void
ContainerDeselectAll(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerDeselectAll", params, num_params));
}

static void
ContainerEndExtend(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerEndExtend", params, num_params));
}

static void
ContainerEndSelect(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerEndSelect", params, num_params));

	/* FIX ME this comes not from documentation but it's a hack */
	ContainerActivate(w, event, params, num_params);
}

static void
ContainerEndToggle(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerEndToggle", params, num_params));
}

static void
ContainerEndTransfer(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerEndTransfer", params, num_params));
}

static void
ContainerExpandOrCollapse(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerExpandOrCollapse", params, num_params));
}

static void
ContainerExtend(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerExtend", params, num_params));
}

static void
ContainerExtendCursor(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerExtendCursor", params, num_params));
}

static void
ContainerHandleBtn1Down(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerHandleBtn1Down", params, num_params));

	/* FIX ME - Do something depending on the state of XmNenableBtn1Transfer */

	/* If the pointer is over an unselected item or background, select it */

	/* Start transfer */

	/* FIX ME - somehow I think we must trigger this action */
	if (num_params) {
		XtCallActionProc(w, params[0], event, params + 1, (*num_params)-1);
	}
}

static void
ContainerHandleBtn1Motion(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerHandleBtn1Motion", params, num_params));
}

/*
 * 
 */
static void
ContainerHandleBtn1Up(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerHandleBtn1Up", params, num_params));

	if (0) {
		/* If a button 1 transfer is in progress, cancel it. */
		/* FIX ME */
	} else {
		/* Perform the action specified in the parameter */
		if (*num_params != 1) {
			_XmWarning(w, "ContainerHandleBtn1Up: #parameters should be 1");
			return;
		}

		XtCallActionProc(w, params[0], event, params, 0);
	}
}

static void
ContainerHandleBtn2Down(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerHandleBtn2Down", params, num_params));
}

static void
ContainerHandleBtn2Motion(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerHandleBtn2Motion", params, num_params));
}

static void
ContainerHandleBtn2Up(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerHandleBtn2Up", params, num_params));
}

static void
ContainerMoveCursor(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerMoveCursor", params, num_params));
}

static void
ContainerPrimaryCopy(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerPrimaryCopy", params, num_params));
}

static void
ContainerPrimaryLink(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerPrimaryLink", params, num_params));
}

static void
ContainerPrimaryMove(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerPrimaryMove", params, num_params));
}

static void
ContainerSelect(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerSelect", params, num_params));
}

static void
ContainerSelectAll(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerSelectAll", params, num_params));
}

static void
ContainerStartTransfer(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerStartTransfer", params, num_params));
}

static void
ContainerToggleMode(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	DEBUGOUT(_LtDebugAction(__FILE__, w, "ContainerToggleMode", params, num_params));
}
